;
; pentaveega.asm
;
; Simple demo for attiny5 6-pin MCU. Drives 8-color VGA out from 3 pins.
; Made to Hackaday 1kB challenge.
;
; Author : Jari Tulilahti


.include "tn5def.inc"

.def temp	= r16
.def zero	= r17
.def vline_lo	= r18
.def vline_hi	= r19
.def pixel	= r20
.def counter	= r21
.def offset	= r22
.def one	= r23

; HSYNC macros
;
.macro hsync_on
	sbi DDRB, PB0
.endm

.macro hsync_off
	cbi DDRB, PB0
.endm

; VSYNC macros
;
.macro vsync_on
	sbi DDRB, PB2
.endm

.macro vsync_off
	cbi DDRB, PB2
.endm

; Mode macros
;
.macro mode_sync
	sbic PORTB, 0
	out DDRB, zero
	out PORTB, zero
.endm

.macro mode_pixel
	ldi temp, 0xFF
	out PORTB, temp
	out DDRB, zero
.endm

; ============
; Reset vector
; ============
;
.org 0x0000
	rjmp reset

; =======================================
; HSYNC routine. This is "the" interrupt.
; =======================================
;
; We should be at the END of horizontal line here,
; we fix the jitter caused by AVR instructions taking
; 1-3 clocks during front porch. Then we pulse the HSYNC,
; increase the vertical counter in registers and then
; check if we're in Vertical blanking area (and jump to
; it's own routine), and finally draw "pixels"
;
; (TIM0_COMPA vector)
;
.org 0x0005
	; First, let's fix jitter
	;
	in temp, TCNT0L
jitterfix:
	ldi ZH, high(jitterfix)
	ldi ZL, low(jitterfix)
	add ZL, temp
	ijmp
	nop

	; Stack pointer
	;
	ldi temp, low(RAMEND)
	out SPL, temp

	; Sync mode, HSYNC on
	;
	mode_sync

	ldi temp, 2
front_porch:
	dec temp
	brne front_porch

	hsync_on

	; Sync pulse. Wait.
	;
	ldi temp, 15
hsync_loop:
	dec temp
	brne hsync_loop

	; HSYNC off
	;
	hsync_off

	; Increase vertical counter, check for blanking area
	;
	add vline_lo, one
	adc vline_hi, zero

	; Check if we are in visible part of screen
	;
	sbrc vline_hi, 1		; Skip jump if bit 9 of vline is clear (<512)
	rjmp vertical_blank		; Jump if we're outside visible screen

	; Back porch
	;
	ldi temp, 5
back_porch:
	dec temp
	brne back_porch

	; Visible screen - Draw here
	;
	mode_pixel

	; Background bars
	;
	out DDRB, pixel

	; Wait for more center of screen
	;
	ldi temp, 28
prewait:
	dec temp
	brne prewait

	; Fetch curvy bar offset
	;
	mov temp, counter
	inc counter
	lsr temp
	lsr temp
	lsr temp
	andi temp, 31
	ldi ZL, low((wave << 1) + 0x4000)
	ldi ZH, high((wave << 1) + 0x4000)
	add ZL, temp
	ld temp, Z

	; Wait for horizontal offset
	;
pixelwait:
	dec temp
	brne pixelwait

	; Draw the bar
	;
	ldi temp, 13
	out DDRB, temp
	nop

	mov temp, vline_lo
	lsr temp
	or temp, pixel
	out DDRB, temp

	; Bar width
	;
	ldi temp, 8
barwait:
	dec temp
	brne barwait

	ldi temp, 13
	out DDRB, temp
	nop
	nop
	nop
	
	; Background bars
	;
	out DDRB, pixel

	; Calculate background stripes
	;
red:
	lds YL, SRAM_START
	lds YH, SRAM_START + 1
	sbrc pixel, 0
	rjmp red_end
	cp vline_lo, YL
	cpc vline_hi, YH
	brne green
	sbr pixel, 1
	rjmp green

red_end:
	ldi temp, 30
	add YL, temp
	adc YH, zero
	cp vline_lo, YL
	cpc vline_hi, YH
	brne green
	cbr pixel, 1

green:
	lds YL, SRAM_START + 2
	lds YH, SRAM_START + 3
	sbrc pixel, 2
	rjmp green_end
	cp vline_lo, YL
	cpc vline_hi, YH
	brne blue
	sbr pixel, 4
	rjmp blue

green_end:
	ldi temp, 30
	add YL, temp
	adc YH, zero
	cp vline_lo, YL
	cpc vline_hi, YH
	brne blue
	cbr pixel, 4

blue:
	lds YL, SRAM_START + 4
	lds YH, SRAM_START + 5
	sbrc pixel, 3
	rjmp blue_end
	cp vline_lo, YL
	cpc vline_hi, YH
	brne color_end
	sbr pixel, 8
	rjmp color_end

blue_end:
	ldi temp, 30
	add YL, temp
	adc YH, zero
	cp vline_lo, YL
	cpc vline_hi, YH
	brne color_end
	cbr pixel, 8

color_end:
	; Wait for next HSYNC
	;
	rjmp wait_hsync


; ==============
; Initialization
; ==============
;
reset:
	; Set up timer interrupt. My counter is not set to 381 (380) because
	; my oscillator is not exactly 12MHz but with 12MHz oscillator, 380
	; is correct value.
	;
	ldi temp, (1 << OCIE0A)			; Compare interrupt enable
	out TIMSK0, temp
	ldi temp, 0x01				; Set counter value.
	out OCR0AH, temp			; 12000000 / 381
	ldi temp, 0x82				; gives vertical refresh
	out OCR0AL, temp			; of about 31.5kHz
	ldi temp, (1 << WGM02) | (1 << CS00)	; CTC mode, prescaler clk/1
	out TCCR0B, temp

	; Set MCU to use external oscillator
	;
	ldi temp, 0xD8
	out CCP, temp
	ldi temp, 0x02
	out CLKMSR, temp
	ldi temp, 0xD8
	out CCP, temp
	out CLKPSR, zero

	;
	; Init stripes
	ldi ZL, low(SRAM_START)
	ldi ZH, high(SRAM_START)
	ldi temp, 100
	st Z+, temp
	st Z+, zero
	ldi temp, 160
	st Z+, temp
	st Z+, zero
	ldi temp, 235
	st Z+, temp
	st Z+, zero

	; Init one
	;
	ldi one, 1

	; Start up
	;
	rjmp wait_hsync

; ======================
; Vertical blanking area
; ======================
;
vertical_blank:
	; Vertical blanking area.
	; Check for VSYNC pulse begin
	;
	cpi vline_lo, 10
	brne check_vp_end		; Nope
	vsync_on			; Yes, just flip VSYNC pin state
	rjmp wait_hsync			; Jump waiting HSYNC

check_vp_end:
	; Check VSYNC pulse end
	;
	cpi vline_lo, 12
	brne check_move_bars		; Not pulse end
	vsync_off			; Flip sync pin state
	rjmp wait_hsync			; Jump waiting HSYNC

check_move_bars:
	; Move bars
	;
	cpi vline_lo, 14
	brne check_screen_end		; No bar movement

	; Move red bar
	;
	lds YL, SRAM_START
	lds YH, SRAM_START + 1
	add YL, one
	adc YH, zero
	ldi temp, 0xD8
	cp YL, temp
	cpc YH, one
	brne store_red
	ldi YL, 0
	ldi YH, 0
store_red:
	sts SRAM_START, YL
	sts SRAM_START + 1, YH

	; Move green bar
	;
	lds YL, SRAM_START + 2
	lds YH, SRAM_START + 3
	sub YL, one
	sbc YH, zero
	cp YL, zero
	cpc YH, zero
	brne store_green
	ldi YL, 0xD8
	ldi YH, 1
store_green:
	sts SRAM_START + 2, YL
	sts SRAM_START + 3, YH


	rjmp wait_hsync			; Jump waiting HSYNC


check_screen_end:
	cpi vline_lo, 45		; Check if we've done 525 lines
	brne wait_hsync			; Not yet, jump waiting HSYNC
	ldi vline_lo, 32		; Start from line 32 (512 - 480 = 32)
	clr vline_hi			; Clear high byte

	; Move curve
	;
	inc offset
	mov counter, offset

; HSYNC WAIT:
; - reset stack pointer
; - enable interrupts
; - wait for HSYNC interrupt
;
wait_hsync:
	sei

wait_isr:
	rjmp wait_isr

.org 0x00F0
wave:
	.db 1, 2, 3, 4, 6, 8,11,14,17,20,22,23,24,24,24,24,24,23,22,20,17,14,11, 8, 6, 4, 3, 2, 1, 1, 1, 1
